from typing import TypeVar, assert_never
from pydantic import BaseModel
from sqlalchemy import Select
from unboil_fastapi_query import FilterOperation, SortOrder


__all__ = [
    "apply_query"
]

T = TypeVar("T")

def apply_query(statement: Select[tuple[T]], query: BaseModel):
    dump = query.model_dump(exclude_unset=True)
    for field, value in dump.items():

        if field.startswith("filter."):
            tail = field.removeprefix("filter.")
            attr_key, order_value = tail.rsplit(".", 1)
            operation = FilterOperation(order_value)
            column = statement.selected_columns.get(attr_key)
            if column is None:
                continue
            if operation == FilterOperation.EQ:
                statement = statement.where(column == value)
            elif operation == FilterOperation.NEQ:
                statement = statement.where(column != value)
            elif operation == FilterOperation.LT:
                statement = statement.where(column < value)
            elif operation == FilterOperation.GT:
                statement = statement.where(column > value)
            elif operation == FilterOperation.LTE:
                statement = statement.where(column <= value)
            elif operation == FilterOperation.GTE:
                statement = statement.where(column >= value)
            elif operation == FilterOperation.LIKE:
                statement = statement.where(column.like(f"%{value}%"))
            elif operation == FilterOperation.ILIKE:
                statement = statement.where(column.ilike(f"%{value}%"))
            elif operation == FilterOperation.STARTSWITH:
                statement = statement.where(column.startswith(value))
            elif operation == FilterOperation.ENDSWITH:
                statement = statement.where(column.endswith(value))
            elif operation == FilterOperation.IN:
                if None not in value:
                    statement = statement.where(column.in_(value))
                else:
                    statement = statement.where(column.in_(value) | column.is_(None))
            elif operation == FilterOperation.NOT_IN:
                statement = statement.where(
                    column.notin_([v for v in value if v is not None])
                )
                if None in value:
                    statement = statement.where(column.isnot(None))
            else:
                assert_never(operation)

        elif field == "sort":
            assert isinstance(value, list) and all(
                isinstance(v, str) for v in value
            ), "Sort must be a list of strings"
            for sort_value in value:
                sort_value: str = sort_value
                attr_key, order_value = sort_value.rsplit(".", 1)
                sort_order = SortOrder(order_value)
                column = statement.selected_columns.get(attr_key)
                if column is None:
                    continue
                if sort_order == SortOrder.ASC:
                    statement = statement.order_by(column.asc())
                elif sort_order == SortOrder.DESC:
                    statement = statement.order_by(column.desc())
                else:
                    assert_never(sort_order)

    return statement



# Helper to test filter operations
from sqlalchemy.orm import InstrumentedAttribute, Session
from unboil_fastapi_query import FilterOperation
from typing import Any, Sequence

import pytest
from unboil_fastapi_query import FilterOperation, build_query
from unboil_fastapi_query.sqlalchemy import apply_query
from sqlalchemy.orm import DeclarativeBase, mapped_column, Session, Mapped
from sqlalchemy import Integer, String, create_engine, select


class Base(DeclarativeBase):
    pass


class Example(Base):
    __tablename__ = "example"
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str | None] = mapped_column(String, nullable=True)


@pytest.fixture(scope="session")
def session():
    engine = create_engine("sqlite:///:memory:")
    Base.metadata.create_all(engine)
    with Session(engine) as session:
        session.add_all(
            [
                Example(id=1, name="Alice"),
                Example(id=2, name="Bob"),
                Example(id=3, name="Alicia"),
                Example(id=4, name=None),
            ]
        )
        session.commit()
        yield session


def filter_test(
    session: Session,
    column: InstrumentedAttribute,
    operation: FilterOperation,
    value: Any,
    expected_names: Sequence[str | None],
):
    QueryModel = build_query(filters={column: [operation]})
    request = QueryModel(**{f"filter.{column.key}.{operation.value}": value})
    stmt = select(Example)
    stmt = apply_query(stmt, request)
    result = session.execute(stmt).scalars().all()
    assert {u.name for u in result} == set(expected_names)


def sort_test(
    session: Session,
    sort_column: InstrumentedAttribute,
    direction: str,
    expected_names: list[str | None],
):
    QueryModel = build_query(sortables=[sort_column])
    request = QueryModel(sort=[f"sort.{sort_column.key}.{direction}"])
    stmt = select(Example)
    stmt = apply_query(stmt, request)
    result = session.execute(stmt).scalars().all()
    names = [u.name for u in result]
    assert names == expected_names


def test_apply_query_eq(session: Session):
    filter_test(session, Example.name, FilterOperation.EQ, "Alice", ["Alice"])


def test_apply_query_eq_none(session: Session):
    filter_test(session, Example.name, FilterOperation.EQ, None, [None])


def test_apply_query_neq(session: Session):
    filter_test(session, Example.name, FilterOperation.NEQ, "Alice", ["Bob", "Alicia"])


def test_apply_query_neq_none(session: Session):
    filter_test(
        session, Example.name, FilterOperation.NEQ, None, ["Alice", "Bob", "Alicia"]
    )


def test_apply_query_lt(session: Session):
    filter_test(session, Example.id, FilterOperation.LT, 2, ["Alice"])


def test_apply_query_gt(session: Session):
    filter_test(session, Example.id, FilterOperation.GT, 2, ["Alicia", None])


def test_apply_query_lte(session: Session):
    filter_test(session, Example.id, FilterOperation.LTE, 2, ["Alice", "Bob"])


def test_apply_query_gte(session: Session):
    filter_test(session, Example.id, FilterOperation.GTE, 2, ["Bob", "Alicia", None])


def test_apply_query_like(session: Session):
    filter_test(session, Example.name, FilterOperation.LIKE, "Ali", ["Alice", "Alicia"])


def test_apply_query_ilike(session: Session):
    filter_test(
        session, Example.name, FilterOperation.ILIKE, "ali", ["Alice", "Alicia"]
    )


def test_apply_query_startswith(session: Session):
    filter_test(
        session, Example.name, FilterOperation.STARTSWITH, "Ali", ["Alice", "Alicia"]
    )


def test_apply_query_endswith(session: Session):
    filter_test(session, Example.name, FilterOperation.ENDSWITH, "cia", ["Alicia"])


def test_apply_query_in(session: Session):
    filter_test(
        session, Example.name, FilterOperation.IN, ["Alice", None], ["Alice", None]
    )


def test_apply_query_not_in(session: Session):
    filter_test(
        session,
        Example.name,
        FilterOperation.NOT_IN,
        ["Alice", None],
        ["Alicia", "Bob"],
    )


def test_sort_by_name_asc(session: Session):
    sort_test(session, Example.name, "asc", [None, "Alice", "Alicia", "Bob"])


def test_sort_by_name_desc(session: Session):
    sort_test(session, Example.name, "desc", ["Bob", "Alicia", "Alice", None])



from enum import StrEnum
from typing import Literal, Optional
from fastapi import Depends, Query
from pydantic import BaseModel, Field, create_model
from sqlalchemy.orm import InstrumentedAttribute, ColumnProperty

__all__ = [
    "QueryDepends",
    "FilterOperation",
    "SortOrder",
    "build_query",
]

def QueryDepends(model: type[BaseModel]) -> BaseModel:
    def f(query = Query()):
        return query
    f.__annotations__["query"] = model 
    return Depends(f)

class FilterOperation(StrEnum):
    EQ = "eq"
    NEQ = "neq"
    LT = "lt"
    GT = "gt"
    LTE = "lte"
    GTE = "gte"
    LIKE = "like"
    ILIKE = "ilike"
    STARTSWITH = "startswith"
    ENDSWITH = "endswith"
    IN = "in"
    NOT_IN = "not_in"


class SortOrder(StrEnum):
    ASC = "asc"
    DESC = "desc"


def build_query(
    filters: dict[InstrumentedAttribute, Literal["*"] | FilterOperation | list[FilterOperation]] = {},
    sortables: list[InstrumentedAttribute] = [],
):

    field_definitions = {}

    # add filters
    for attr, operations in filters.items():
        
        if operations == "*":
            operations = [op for op in FilterOperation]
        elif isinstance(operations, FilterOperation):
            operations = [operations]
            
        for operation in operations:
            if isinstance(attr.property, ColumnProperty):
                field_name = f"filter.{attr.key}.{operation.value}"
                if operation in [FilterOperation.IN, FilterOperation.NOT_IN]:
                    field_definitions[field_name] = (
                        list[Optional[attr.property.columns[0].type.python_type]],
                        Field(default=None),
                    )
                else:
                    field_definitions[field_name] = (
                        Optional[attr.property.columns[0].type.python_type],
                        Field(default=None),
                    )

    # add sorting
    options: list[str] = []
    for attr in sortables:
        options.append(f"{attr.key}.{SortOrder.ASC.value}")
        options.append(f"{attr.key}.{SortOrder.DESC.value}")
    if options:
        field_definitions["sort"] = (
            list[Literal[*options]],
            Field(default_factory=list, json_schema_extra={"uniqueItems": True}),
        )

    return create_model("Query", **field_definitions)
    


from fastapi import FastAPI
from sqlalchemy import create_engine, select
from sqlalchemy.orm import Session


from .schemas import ExampleCreate, ExampleRead
from .models import Base, Example
from unboil_fastapi_query import FilterOperation, QueryDepends, build_query
from unboil_fastapi_query.sqlalchemy import apply_query


engine = create_engine("sqlite:///db.sqlite", echo=False)
Base.metadata.create_all(engine)

app = FastAPI()

@app.post("/", response_model=ExampleRead)
def create(request: ExampleCreate):
    with Session(engine, expire_on_commit=False) as session:
        example = Example(**request.model_dump(exclude_unset=True))
        session.add(example)
        session.commit()
        return example

ExampleQuery = build_query(
    filters={
        Example.id: FilterOperation.EQ,
        Example.name: "*",
    },
    sortables=[
        Example.name,
    ]
)

@app.get("/", response_model=list[ExampleRead])
def list(
    query = QueryDepends(ExampleQuery)
):
    with Session(engine) as session:
        stmt = select(Example)
        stmt = apply_query(stmt, query)
        result = session.execute(stmt)
        return result.scalars().all()


from uuid import uuid4
from sqlalchemy import String
from sqlalchemy.orm import DeclarativeBase, mapped_column, Mapped

class Base(DeclarativeBase):
    pass

class Example(Base):
    __tablename__ = "examples"
    id: Mapped[str] = mapped_column(String, primary_key=True, default=lambda: uuid4().hex)
    name: Mapped[str] = mapped_column(String)
